// Double squarawave music with PWM played on the STAF.
// wav.c
// This program generates a .wav file which sounds the same as what the STAF
// will play.
// 
// 15.08.2016
// Copyright (C) 2014-2016  Balthasar Szczepański
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
	#include <stdio.h>
	#include "definitions.h"
	#include "sheet.h"
	
	#define RTCSTOP 1465
	#define WDTSTOP  128
	// #define SPRSTOP  136
	// #define SPR    44100
	#define SPRSTOP  128
	#define SPR    46875
	
	int main( int argc, char *argv[] )
	{
		FILE *wavfile;
		
		unsigned char channel; 
		unsigned char active=0;
		unsigned char select=0x33;
		
		unsigned short wavestop1=0;
		unsigned short wavestop2=0;
		unsigned short wavecount1=0;
		unsigned short wavecount2=0;
		
		unsigned short noteend1=0;
		unsigned short noteend2=0;
		unsigned short notestop1=0;
		unsigned short notestop2=0;
		unsigned short notecount1=0xf000;
		unsigned short notecount2=0xf000;
		
		unsigned short sheetcount1=0;
		unsigned short sheetcount2=0;
		
		unsigned short lengthoffset1=0;
		unsigned short lengthoffset2=0;
		
		unsigned char sample=0;
		
		unsigned char stop1=0;
		unsigned char stop2=0;
		
		unsigned short rtccount=0xf000;
		unsigned short wdtcount=0xf000;
		unsigned short sprcount=0xf000;
		unsigned long  samples=0;
		
		unsigned long  size1pos;
		unsigned long  size2pos;
		
		unsigned long  templong;
		unsigned short tempshort;
		
		if(argc < 2)
		{
			fputs("Output file path missing.\n",stderr);
			return 2;
		}
		
		wavfile = fopen(argv[1],"wb");
		if (wavfile==NULL)
		{
			fputs("Can,t write .wav file.\n",stderr);
			return 1;
		}
		
		//RIFF HEADER
		fwrite("RIFF",1,4,wavfile); //ChunkID
		size1pos=ftell(wavfile);//correct this later
		templong=36;
		fwrite(&templong,4,1,wavfile); //ChunkSize
		fwrite("WAVE",1,4,wavfile); //Format
		
		//fmt subchunk
		fwrite("fmt ",1,4,wavfile); //Subchunk1ID
		templong=16;
		fwrite(&templong,4,1,wavfile); //Subchunk1Size
		tempshort=1;
		fwrite(&tempshort,2,1,wavfile); //AudioFormat
		fwrite(&tempshort,2,1,wavfile); //NumChannels
		templong=SPR;
		fwrite(&templong,4,1,wavfile); //SampleRate
		fwrite(&templong,4,1,wavfile); //ByteRate
		tempshort=1;
		fwrite(&tempshort,2,1,wavfile); //BlockAlign
		tempshort=8;
		fwrite(&tempshort,2,1,wavfile); //BitsPerSample
		
		//data subchunk
		fwrite("data",1,4,wavfile); //Subchunk2ID
		templong=0;
		size2pos=ftell(wavfile);//correct this later
		fwrite(&templong,4,1,wavfile); //Subchunk2Size
		
		
		//each iteration is one tick of a 6MHz clock.
		//Both channels must end with a STOP, otherwise this loops forever.
		while(!stop1 || !stop2)
		{
			++rtccount;
			++wdtcount;
			++sprcount;
			
			//equivalent of interrupt from RTC
			if(rtccount>=RTCSTOP)
			{
				rtccount=0;
				++notecount1;
				++notecount2;
				if(notecount1>=notestop1)
				{
					active &= ~0x01;
					notestop1=0xffff;
				}
				
				if(notecount1>=noteend1)
				{
					notecount1=0;
					fprintf(stderr,"1: %hu\n",sheetcount1);
					while(sheet1[sheetcount1]>= NXSP)
					{
						switch(sheet1[sheetcount1]& LENG)
						{
						case NXSP:
							lengthoffset1+=10;
							++sheetcount1;
							break;
						case STOP:
							stop1=1;
						case REPL:
							lengthoffset1=0;
							sheetcount1 = 0;
							noteend1 = 0;
							active &= ~0x01;
							break;
						default:
							++sheetcount1;
						}
					}
					if(!stop1)
					{
						noteend1=length[((sheet1[sheetcount1] & LENG)>>12)+lengthoffset1];
						if (sheet1[sheetcount1] & PTCH)
						{
							wavestop1=sheet1[sheetcount1] & PTCH;
							if(wavestop1==wavestop2)
							{
								wavecount1=wavecount2 + (wavestop2>>1);
								if(wavecount1 > wavestop2)
									wavecount1-=wavestop2;
							}
							active |= 0x01;
							if (sheet1[sheetcount1] & SNGL)
								notestop1=noteend1-STOP_OFFSET;
							else
								notestop1=0xffff;
						}
						else
						{
							active &= ~0x01;
							notestop1=0xffff;
						}
						++sheetcount1;
					}
				}
				
				if(notecount2>=notestop2)
				{
					active &= ~0x02;
					notestop2=0xffff;
				}
				
				if(notecount2>=noteend2)
				{
					notecount2=0;
					fprintf(stderr,"2: %hu\n",sheetcount2);
					while(sheet2[sheetcount2]>= NXSP)
					{
						switch(sheet2[sheetcount2]& LENG)
						{
						case NXSP:
							lengthoffset2+=10;
							++sheetcount2;
							break;
						case STOP:
							stop2=1;
						case REPL:
							lengthoffset2=0;
							sheetcount2 = 0;
							noteend2 = 0;
							active &= ~0x02;
							break;
						default:
							++sheetcount2;
						}
					}
					if(!stop2)
					{
						noteend2=length[((sheet2[sheetcount2] & LENG)>>12)+lengthoffset2];
						if (sheet2[sheetcount2] & PTCH)
						{
							wavestop2=sheet2[sheetcount2] & PTCH;
							if(wavestop2==wavestop1)
							{
								wavecount2=wavecount1 + (wavestop1>>1);
								if(wavecount2 > wavestop1)
									wavecount2-=wavestop1;
								// channel=(channel&0x01)?0x03:0x00;
							}
							active |= 0x02;
							if (sheet2[sheetcount2] & SNGL)
								notestop2=noteend2-STOP_OFFSET;
							else
								notestop2=0xffff;
						}
						else
						{
							active &= ~0x02;
							notestop2=0xffff;
						}
						++sheetcount2;
					}
				}
			}
			
			//equivalent of interrupt from WDT
			if(wdtcount>=WDTSTOP)
			{
				wdtcount=0;
				++wavecount1;
				++wavecount2;
				if(wavecount1>wavestop1)
				{
					wavecount1=0;
					channel^=0x01;
				}
				if(wavecount2>wavestop2)
				{
					wavecount2=0;
					channel^=0x02;
				}
				active&=select;
				channel&=active;
				if(channel==0x00)
					sample = 0x20;
				else if(channel==0x03)
					sample=0xe0;
				else
					sample=0x80;
			}
			
			//time to add next sample
			if(sprcount>=SPRSTOP)
			{
				sprcount=0;
				fwrite(&sample,1,1,wavfile);
				++samples;
			}
		}
		
		fseek(wavfile,size2pos,SEEK_SET);
		fwrite(&samples,4,1,wavfile); //Subchunk2Size
		samples+=36;
		fseek(wavfile,size1pos,SEEK_SET);
		fwrite(&samples,4,1,wavfile); //ChunkSize
		
		
		
		if(argc >= 1)
		{
			fclose(wavfile);
		}
		return 0;
	}
